﻿// COPYRIGHT 2010, 2012 by the Open Rails project.
// 
// This file is part of Open Rails.
// 
// Open Rails is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// Open Rails is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Open Rails.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using ORTS.Scripting.Api;

namespace ORTS.Scripting.Script
{
    public class PBL2BrakeController : BrakeController
    {
        enum State
        {
            Overload,
            QuickRelease,
            Release,
            Lap,
            Apply,
            Emergency
        }

        public float OverloadValue { get; private set; }
        public float QuickReleaseValue { get; private set; }
        public float ReleaseValue { get; private set; }
        public float LapValue { get; private set; }
        public float ApplyValue { get; private set; }
        public float EmergencyValue { get; private set; }

        // brake controller values
        private float OverloadPressurePSI = 5.8F;
        private float FirstDepressurePSI = 4.35F;
        private float BrakeReleasedDepressurePSI = 2.9F;

        protected float CurrentPressure;

        private State CurrentState;
        private State PreviousState;

        private bool FirstDepressure = false;
        private bool Neutral = false;
        private bool Overload = false;
        private bool QuickRelease = false;
        private bool Release = false;
        private bool Apply = false;

        public PBL2BrakeController()
        {
        }

        public override void Initialize()
        {
            foreach (MSTSNotch notch in Notches())
            {
                switch (notch.Type)
                {
                    case MSTSNotchType.Release:
                        ReleaseValue = notch.Value;
                        break;
                    case MSTSNotchType.FullQuickRelease:
                        OverloadValue = notch.Value;
                        QuickReleaseValue = notch.Value;
                        break;
                    case MSTSNotchType.Lap:
                        LapValue = notch.Value;
                        break;
                    case MSTSNotchType.Apply:
                    case MSTSNotchType.GSelfLap:
                    case MSTSNotchType.GSelfLapH:
                        ApplyValue = notch.Value;
                        break;
                    case MSTSNotchType.Emergency:
                        EmergencyValue = notch.Value;
                        break;
                }
            }
        }

        public override float Update(float elapsedSeconds)
        {
            if (Apply)
                SetCurrentValue(ApplyValue);
            else if (Release)
                SetCurrentValue(ReleaseValue);
            else
                SetCurrentValue(LapValue);

            return CurrentValue();
        }

        public override void UpdatePressure(ref float pressurePSI, float elapsedClockSeconds, ref float epPressurePSI)
        {
            if (!FirstDepressure && Apply && pressurePSI > MaxPressurePSI() - FirstDepressurePSI)
                FirstDepressure = true;
            else if (FirstDepressure && pressurePSI <= MaxPressurePSI() - FirstDepressurePSI)
                FirstDepressure = false;

            if (Apply && Overload)
                Overload = false;
            if (Apply && QuickRelease)
                QuickRelease = false;

            if (EmergencyBraking() || TCSEmergencyBraking())
                CurrentState = State.Emergency;
            else if (
                Apply && pressurePSI > MaxPressurePSI() - FullServReductionPSI()
                || !Overload && pressurePSI > MaxPressurePSI()
                || FirstDepressure && !Release && !QuickRelease && pressurePSI > MaxPressurePSI() - FirstDepressurePSI
                )
                CurrentState = State.Apply;
            else if (Overload && pressurePSI < MaxPressurePSI() + OverloadPressurePSI)
                CurrentState = State.Overload;
            else if (QuickRelease && !Neutral && pressurePSI < MaxPressurePSI())
                CurrentState = State.QuickRelease;
            else if (
                !Neutral && (
                    Release && pressurePSI < MaxPressurePSI()
                    || !FirstDepressure && pressurePSI > MaxPressurePSI() - BrakeReleasedDepressurePSI && pressurePSI < MaxPressurePSI()
                    || pressurePSI < MaxPressurePSI() - FullServReductionPSI()
                    )
                )
                CurrentState = State.Release;
            else
                CurrentState = State.Lap;

            switch (CurrentState)
            {
                case State.Overload:
                    SetUpdateValue(-1);

                    pressurePSI += QuickReleaseRatePSIpS() * elapsedClockSeconds;
                    epPressurePSI -= QuickReleaseRatePSIpS() * elapsedClockSeconds;

                    if (pressurePSI > MaxPressurePSI() + OverloadPressurePSI)
                        pressurePSI = MaxPressurePSI() + OverloadPressurePSI;
                    break;

                case State.QuickRelease:
                    SetUpdateValue(-1);

                    pressurePSI += QuickReleaseRatePSIpS() * elapsedClockSeconds;
                    epPressurePSI -= QuickReleaseRatePSIpS() * elapsedClockSeconds;

                    if (pressurePSI > MaxPressurePSI())
                        pressurePSI = MaxPressurePSI();
                    break;

                case State.Release:
                    SetUpdateValue(-1);

                    pressurePSI += ReleaseRatePSIpS() * elapsedClockSeconds;
                    epPressurePSI -= ReleaseRatePSIpS() * elapsedClockSeconds;

                    if (pressurePSI > MaxPressurePSI())
                        pressurePSI = MaxPressurePSI();
                    break;

                case State.Lap:
                    SetUpdateValue(0);
                    break;

                case State.Apply:
                    SetUpdateValue(1);

                    pressurePSI -= ApplyRatePSIpS() * elapsedClockSeconds;
                    epPressurePSI += ApplyRatePSIpS() * elapsedClockSeconds;

                    if (pressurePSI < MaxPressurePSI() - FullServReductionPSI())
                        pressurePSI = MaxPressurePSI() - FullServReductionPSI();
                    break;

                case State.Emergency:
                    SetUpdateValue(1);

                    pressurePSI -= EmergencyRatePSIpS() * elapsedClockSeconds;

                    if (pressurePSI < 0)
                        pressurePSI = 0;
                    break;
            }

            if (epPressurePSI > MaxPressurePSI())
                epPressurePSI = MaxPressurePSI();
            if (epPressurePSI < 0)
                epPressurePSI = 0;

            if (QuickRelease && pressurePSI == MaxPressurePSI())
                QuickRelease = false;

            PreviousState = CurrentState;
        }

        public override void UpdateEngineBrakePressure(ref float pressurePSI, float elapsedClockSeconds)
        {
            switch (CurrentState)
            {
                case State.Release:
                    SetCurrentValue(ReleaseValue);
                    pressurePSI -= ReleaseRatePSIpS() * elapsedClockSeconds;
                    break;
                
                case State.Apply:
                    SetCurrentValue(ApplyValue);
                    pressurePSI += ApplyRatePSIpS() * elapsedClockSeconds;
                    break;
                
                case State.Emergency:
                    SetCurrentValue(EmergencyValue);
                    pressurePSI += EmergencyRatePSIpS() * elapsedClockSeconds;
                    break;
            }

            if (CurrentState > PreviousState)
                SetUpdateValue(1);
            else if (CurrentState < PreviousState)
                SetUpdateValue(-1);
            else
                SetUpdateValue(0);


            if (pressurePSI > MaxPressurePSI())
                pressurePSI = MaxPressurePSI();
            if (pressurePSI < 0)
                pressurePSI = 0;

            PreviousState = CurrentState;
        }

        public override void HandleEvent(BrakeControllerEvent evt)
        {
            switch (evt)
            {
                case BrakeControllerEvent.StartIncrease:
                    Apply = true;
                    break;

                case BrakeControllerEvent.StopIncrease:
                    Apply = false;
                    break;

                case BrakeControllerEvent.StartDecrease:
                    Release = true;
                    break;

                case BrakeControllerEvent.StopDecrease:
                    Release = false;
                    break;
            }
        }

        public override void HandleEvent(BrakeControllerEvent evt, float? value)
        {
            switch (evt)
            {
                case BrakeControllerEvent.StartIncrease:
                    Apply = true;
                    break;

                case BrakeControllerEvent.StartDecrease:
                    Release = true;
                    break;

                case BrakeControllerEvent.SetRDPercent:
                    if (value != null)
                    {
                        float newValue = value ?? 0F;
                        SetRDPercent(newValue);
                    }
                    break;

                case BrakeControllerEvent.SetCurrentValue:
                    if (value != null)
                    {
                        float newValue = value ?? 0F;
                        SetValue(newValue);
                    }
                    break;
            }
        }

        public override bool IsValid()
        {
            return true;
        }

        public override string GetStatus()
        {
            switch (CurrentState)
            {
                case State.Overload:
                    return "Overload";

                case State.QuickRelease:
                    return "Quick Release";

                case State.Release:
                    return "Release";

                case State.Lap:
                    return "Lap";

                case State.Apply:
                    return "Apply";

                case State.Emergency:
                    if (EmergencyBraking())
                        return "Emergency Braking Push Button";
                    else if (TCSEmergencyBraking())
                        return "TCS Emergency Braking";
                    else if (TCSFullServiceBraking())
                        return "TCS Full Service Braking";
                    else
                        return "Emergency";

                default:
                    return "";
            }
        }

        private float SetRDPercent(float percent)
        {
            if (percent < 40)
            {
                Apply = true;
                Release = false;
            }
            else if (percent > 60)
            {
                Apply = false;
                Release = true;
            }
            else
            {
                Apply = false;
                Release = false;
            }

            return CurrentValue() * 100;
        }

        private void SetValue(float v)
        {
            SetCurrentValue(v);
            
            if (CurrentValue() == EmergencyValue)
            {
                Apply = false;
                Release = false;
                QuickRelease = false;
            }
            else if (CurrentValue() == ApplyValue)
            {
                Apply = true;
                Release = false;
                QuickRelease = false;
            }
            else if (CurrentValue() == LapValue)
            {
                Apply = false;
                Release = false;
                QuickRelease = false;
            }
            else if (CurrentValue() == ReleaseValue)
            {
                Apply = false;
                Release = true;
                QuickRelease = false;
            }
            else if (CurrentValue() == QuickReleaseValue)
            {
                Apply = false;
                Release = false;
                QuickRelease = true;
            }
        }
    }
}
